{**********************************************************************

	MDI Program Support Objects		$Version$
        $Author$			$Date$

        Copyright (c) 1991 Anthony M. Vitabile.
        All rights reserved.

        Permission is hereby granted to all who desire to use this
        source code in their programs provided the above copyright
        is included in the program.

        Unit Description

        This unit implements three basic object types for use in MDI
        programs under Windows.  These are:

        TMDIChild:  A normal MDI child window which supports the use of
        a special menu whenever the window is active;

        TMDIDialog:  An MDI window which contains a modeless dialog box
        as its only child.  The window adjusts itself so that it looks
        like the non-client portion of a dialog box.

        TMDIFrame:  A frame window type descended from TMDIWindow that
        supports locating and closing a particular child window when
        that child window's handle is provided to it.  This can serve
        as the basis for a method that closes the currently active
        MDI child window.

**********************************************************************}

UNIT MDITypes;
INTERFACE
  USES WinTypes, Objects, ODialogs, OWindows;

  TYPE
    pMDIChild  = ^tMDIChild;
    pMDIDialog = ^tMDIDialog;
    pMDIFrame  = ^tMDIFrame;

    tMDIChild  = OBJECT (tWindow)
                   MenuPos:  INTEGER;
                   Menu   ,
                   WndMenu:  hWND;

                   CONSTRUCTOR Init(aParent :  pWindowsObject;
                                    aTitle  :  pChar;
                                    aMenu   :  hMenu;
                                    aMenuPos:  INTEGER);
                   CONSTRUCTOR Load(VAR S:  tStream);
                   PROCEDURE Store(VAR S:  tStream);
                   PROCEDURE wmMDIActivate(VAR Msg:  tMessage);
                     VIRTUAL wm_First + wm_MDIActivate;
                 END;

    tMDIDialog = OBJECT (tMDIChild)
                   SysMenu  :  hMenu;
                   NCHeight :  INTEGER;
                   TheDialog:  pDlgWindow;

                   PROCEDURE ActivateDlg(VAR Msg:  tMessage);
                   DESTRUCTOR Done;
                     VIRTUAL;
                   CONSTRUCTOR Init(aParent :  pWindowsObject;
                                    aTitle  :  pChar;
                                    aMenu   :  hMenu;
                                    aMenuPos:  INTEGER);
                   FUNCTION InitDialog:  pDlgWindow;
                     VIRTUAL;
                   CONSTRUCTOR Load(VAR S:  tStream);
                   PROCEDURE Store(VAR S:  tStream);
                   PROCEDURE SetChildFocus(VAR Msg:  tMessage);
                   PROCEDURE SetupWindow;
                     VIRTUAL;
                   PROCEDURE wmActivate(VAR Msg:  tMessage);
                     VIRTUAL wm_First + wm_Activate;
                   PROCEDURE wmChildActivate(VAR Msg:  tMessage);
                     VIRTUAL wm_First + wm_ChildActivate;
                   PROCEDURE wmInitMenu(VAR Msg:  tMessage);
                     VIRTUAL wm_First + wm_InitMenu;
                   PROCEDURE wmGetText(VAR Msg:  tMessage);
                     VIRTUAL wm_First + wm_GetText;
                   PROCEDURE wmMDIActivate(VAR Msg:  tMessage);
                     VIRTUAL wm_First + wm_MDIActivate;
                   PROCEDURE wmSetFocus(VAR Msg:  tMessage);
                     VIRTUAL wm_First + wm_SetFocus;
                   PROCEDURE wmSize(VAR Msg:  tMessage);
                     VIRTUAL wm_First + wm_Size;
                   PROCEDURE wmSysCommand(VAR Msg:  tMessage);
                     VIRTUAL wm_First + wm_SysCommand;
                 END;

    tMDIFrame  = OBJECT (tMDIWindow)
                   PROCEDURE CloseChild(aChild:  pWindowsObject);
                     VIRTUAL;
                   FUNCTION GetChild(TheWnd:  hWND):  pWindowsObject;
                     VIRTUAL;
                   PROCEDURE wmInitMenu(VAR Msg:  tMessage);
                     VIRTUAL wm_First + wm_InitMenu;
                   PROCEDURE wmInitMenuPopup(VAR Msg:  tMessage);
                     VIRTUAL wm_First + wm_InitMenuPopup;
                 END;

IMPLEMENTATION
  USES Strings, WinProcs;

  CONST
    RMDIChild :  tStreamRec = (ObjType: 10000;
                               VmtLink: Ofs(TypeOf(tMDIChild)^);
                               Load:    @tMDIChild.Load;
                               Store:   @tMDIChild.Store);

    RMDIDialog:  tStreamRec = (ObjType: 10001;
                               VmtLink: Ofs(TypeOf(tMDIDialog)^);
                               Load:    @tMDIDialog.Load;
                               Store:   @tMDIDialog.Store);

    RMDIFrame :  tStreamRec = (ObjType: 10002;
                               VmtLink: Ofs(TypeOf(tMDIFrame)^);
                               Load:    @tMDIFrame.Load;
                               Store:   @tMDIFrame.Store);

{**********************************************************************

	TMDIChild Methods

	Init Constructor.  Creates & initlializes an instance of
        TMDIChild.  Most of the window's initialization is passed onto
        TWindow, the ancestor object type of this type.  We save the
        child's frame menu handle in the inherited Attr.Menu field, then
        save the handle of the Window popup menu, whose position in the
        menu is given by AMenuPos.

**********************************************************************}

  CONSTRUCTOR tMDIChild.Init(aParent :  pWindowsObject;
                             aTitle  :  pChar;
                             aMenu   :  hMenu;
                             aMenuPos:  INTEGER);
    BEGIN	{ TMDIChild.Init }
      tWindow.Init(aParent, aTitle);
      Menu    := aMenu;
      MenuPos := aMenuPos;
      IF Menu = 0
        THEN WndMenu := 0
        ELSE WndMenu := GetSubMenu(Menu, MenuPos)
    END		{ TMDIChild.Init };

{**********************************************************************

	Load Constructor.  Creates & initlializes an instance of
        TMDIChild that has been read from a stream.

**********************************************************************}

  CONSTRUCTOR tMDIChild.Load(VAR S:  tStream);
    BEGIN	{ TMDIChild.Load }
      tWindow.Load(S);
      S.Read(Menu   , SizeOf(Menu));
      S.Read(MenuPos, SizeOf(MenuPos));
      WndMenu := 0
    END		{ TMDIChild.Load };

{**********************************************************************

	Store method.  Writes an instance of a TMDIChild onto the stream.

**********************************************************************}

  PROCEDURE tMDIChild.Store(VAR S:  tStream);
    BEGIN	{ TMDIChild.Store }
      tWindow.Store(S);
      S.Write(Menu   , SizeOf(Menu));
      S.Write(MenuPos, SizeOf(MenuPos))
    END		{ TMDIChild.Store };

{**********************************************************************

        wmMDIActivate message handler.  This method is executed whenever
        the MDI child window receives a wm_MDIActivate message.  The
        action taken is to reset the menu currently being used by the
        current frame window to either the menu specified by the
        Attr.Menu field of this TMDIChild instance if the window is
        being activated (wParam <> 0), or to reset it to the default
        menu for the frame as specified by the frame's Attr.Menu field.

        Note that in either case, if the handle of the new menu is 0,
        then the current menu is not changed!

**********************************************************************}

  PROCEDURE tMDIChild.wmMDIActivate(VAR Msg:  tMessage);
    VAR
      NewMenu     ,
      NewPos      :  hMenu;
      MDIClient   :  PMDIClient;
      MDIClientWnd:  hWND;

    BEGIN	{ TMDIChild.wmMDIActivate }
      MDIClient    := PMDIWindow(Parent)^.GetClient;
      MDIClientWnd := MDIClient^.hWindow;
      IF Msg.wParam = 0
        THEN
          BEGIN					{ Get frame window menu data  }
            NewMenu := PMDIWindow(Parent)^.Attr.Menu;
            NewPos  := MDIClient^.ClientAttr.hWindowMenu
          END
        ELSE
          BEGIN					{ Get MDI child window menu data }
            NewMenu := Menu;
            NewPos  := WndMenu
          END;					{ Now Set the menu up }
      IF NewMenu <> 0
        THEN
          BEGIN
            SendMessage(MDIClientWnd, wm_MDISetMenu, 0, MakeLong(NewMenu, NewPos));
            DrawMenuBar(Parent^.hWindow)	{ Redraw the menu bar }
          END;
      Msg.Result := 0				{ Return 0 }
    END		{ TMDIChild.wmMDIActivate };

{**********************************************************************

	TMDIDialog methods

        Activate the child dialog box method.  This method is called
        whenever the MDI dialog is activated or receives a message which
        affects the keyboard handling.  If the window is not an icon,
        then the current keyboard handler is set to the window's child
        dialog box.  Otherwise, the keyboard handler is set to nil.

**********************************************************************}

  PROCEDURE tMDIDialog.ActivateDlg(VAR Msg:  tMessage);
    BEGIN	{ TMDIDialog.ActivateDlg }
      IF IsIconic(hWindow)				{ If the window isn't an icon right now }
        THEN Application^.SetKBHandler(NIL)		{ Otherwise turn off the keyboard handler }
        ELSE Application^.SetKBHandler(TheDialog);	{ Turn on the dialog's keyboard handler }
      Msg.Result := 0
    END		{ TMDIDialog.ActivateDlg };

{**********************************************************************

	Destroys a TMDIDialog.  The routine first disposes of the
        child dialog box, then destroys itself.

**********************************************************************}

  DESTRUCTOR tMDIDialog.Done;
    BEGIN	{ TMDIDialog.Done }
      Dispose(TheDialog, Done);
      Menu    := 0;
      WndMenu := 0;
      SysMenu := 0;
      tMDIChild.Done
    END		{ TMDIDialog.Done };

{**********************************************************************

	Init contructor.  This method creates a new TMDIDialog.  It
        allows TMDIChild to do most of the initialization, and then it
        initializes the handle of the window's system menu to 0.  It
        then calls the virtual method InitDialog to actually create
        an OWL TDlgWindow object.

**********************************************************************}

  CONSTRUCTOR tMDIDialog.Init(aParent :  pWindowsObject;
                              aTitle  :  pChar;
                              aMenu   :  hMenu;
                              aMenuPos:  INTEGER);
    BEGIN	{ TMDIDialog.Init }
      tMDIChild.Init(aParent, aTitle, aMenu, aMenuPos);
      SysMenu   := 0;
      TheDialog := InitDialog
    END		{ TMDIDialog.Init };

{**********************************************************************

	InitDialog method.  This method creates a TDlgWindow, which
        is to be the child of the TMDIDialog window.  This is an
        abstract method which must be overridden by a descendant type.

**********************************************************************}

  FUNCTION tMDIDialog.InitDialog:  pDlgWindow;
    BEGIN	{ TMDIDialog.InitDialog }
      abstract
    END		{ TMDIDialog.InitDialog };

{**********************************************************************

	Load Constructor.  Creates & initlializes an instance of
        TMDIDialog that has been read from a stream.

**********************************************************************}

  CONSTRUCTOR tMDIDialog.Load(VAR S:  tStream);
    BEGIN	{ TMDIDialog.Load }
      tMDIChild.Load(S);
      GetChildPtr(S, TheDialog)
    END		{ TMDIChild.Load };

{**********************************************************************

	Store method.  Writes an instance of a TMDIDialog onto the stream.

**********************************************************************}

  PROCEDURE tMDIDialog.Store(VAR S:  tStream);
    BEGIN	{ TMDIDialog.Store }
      tMDIChild.Store(S);
      PutChildPtr(S, TheDialog)
    END		{ TMDIDialog.Store };

{**********************************************************************

	SetChildFocus method.  This method is responsible for saving
        and properly restoring the focus to the TDlgWindow's child
        controls once the TMDIDialog gains or loses activation or
        focus.  The method uses the TMDIDialog's FocusChildHandle
        field, inherited from TWindow, to track the handle of the
        dialog's child that has the focus at the time the window is
        deactivated.

**********************************************************************}

  PROCEDURE tMDIDialog.SetChildFocus(VAR Msg:  tMessage);
    VAR
      CurrentFocus,
      Dlg         :  hWND;

    BEGIN	{ TMDIDialog.SetChildFocus }
      Dlg := TheDialog^.hWindow;			{ Make life easier for us }
      IF NOT IsIconic(hWindow)				{ If the window isn't an icon right now }
        THEN
          IF Msg.wParam <> 0				{ If the window is being activated }
            THEN					{ Set the focus to the correct child handle }
              BEGIN
                IF FocusChildHandle = 0			{ If we haven't got any focus yet }
                  THEN SetFocus(TheDialog^.hWindow)	{ Then set the focus to the dialog }
                  ELSE					{ Set the focus to the correct window }
                    IF IsWindow(FocusChildHandle)	{ Make sure we have a valid handle here! }
                      THEN SetFocus(FocusChildHandle)
              END
            ELSE					{ Save the current focus }
              BEGIN
                CurrentFocus := GetFocus;		{ Get the window that current has focus }
                IF (CurrentFocus <> 0) AND		{ If there is one that has the focus }
                    IsChild(Dlg, CurrentFocus)		{ And it's a child of the dialog }
                  THEN FocusChildHandle := CurrentFocus	{ Then save the child's handle }
              END;
      Msg.Result := 0
    END		{ TMDIDialog.SetChildFocus };

{**********************************************************************

	SetupWindow.  Creates all window elements and then sets the
        focus to the child TDlgWindow dialog box.  It then initializes
        FocusChildHandle to the handle of the window that currently has
        the focus.

**********************************************************************}

  PROCEDURE tMDIDialog.SetupWindow;
    VAR
      h   ,
      w   :  INTEGER;
      Rect:  tRect;

    BEGIN	{ TMDIDialog.SetupWindow }
      tMDIChild.SetupWindow;			{ Do ancestor setup }
      SetFocus(TheDialog^.hWindow);		{ Set the focus to the dialog }
      GetWindowRect(hWindow, Rect);		{ Get the window's rectangle }
      h := Rect.Bottom - Rect.Top;		{ Compute the rectangle's height }
      GetClientRect(hWindow, Rect);		{ Now get the window's client rectangle }
      NCHeight := h - Rect.Bottom - Rect.Top;	{ The correct Non-Client height is the height difference }
      GetClientRect(TheDialog^.hWindow, Rect);	{ Finally, get the dialog box's client rectangle }
      Attr.h := Rect.Bottom - Rect.Top + NCHeight;
      Attr.w := Rect.Right  - Rect.Left
    END		{ TMDIDialog.SetupWindow };


{**********************************************************************

	Track the window's activation state.  When this message is
        received, the window must activate the child dialog and then
        set the child focus.  This message is received when the parent
        application is restored from an icon.  The top most MDI child
        gets the message immediately following the restoration of
        the Frame window.

**********************************************************************}

  PROCEDURE tMDIDialog.wmActivate(VAR Msg:  tMessage);
    BEGIN	{ TMDIDialog.wmActivate }
      ActivateDlg  (Msg);
      SetChildFocus(Msg);
      DefWndProc   (Msg)
    END		{ TMDIDialog.wmActivate };

{**********************************************************************

	We get this message because Windows is playing with the dialog
        window.  This usually happens during processing of wm_Activate
        and wm_MDIActivate.  We activate the child TDlgWindow dialog
        box's keyboard handler after first passing it onto the default
        message procedure for the window.

        NOTE THAT WE LEAVE THE FOCUS ALONE!

**********************************************************************}

  PROCEDURE tMDIDialog.wmChildActivate(VAR Msg:  tMessage);
    BEGIN	{ TMDIDialog.wmChildActivate }
      ActivateDlg(Msg);
      DefWndProc (Msg)
    END		{ TMDIDialog.wmChildActivate };

{**********************************************************************

	This method adds code to the wm_GetText handling that stores
        current dialog box child window control handle.  This is done
        here to handle the case where the current dialog box has the
        focus & control is switched away from the application without
        minimizing it.

        It is *important* that this method call DefWndProc!!!!!!!!

**********************************************************************}

  PROCEDURE tMDIDialog.wmGetText(VAR Msg:  tMessage);
    VAR
      CurFocus:  hWND;

    BEGIN	{ TMDIDialog.wmGetText }
      CurFocus := GetFocus;				{ Get the window that current has focus }
      IF (CurFocus <> 0) AND				{ If there is one that has the focus }
          IsChild(TheDialog^.hWindow, CurFocus)		{ And it's a child of the dialog }
        THEN FocusChildHandle := CurFocus;		{ Then save the child's handle }
      DefWndProc(Msg)
    END		{ TMDIDialog.wmGetText };

{**********************************************************************

	This method handles saving the focus for the current dialog
        child window when a menu item is selected.  This helps to ensure
        that the focus goes to the right control at all times.

        The method also disables the Maximize and Size options on the
        child window's system menu.  This allows the code to prevent
        the window's size from being changed.

**********************************************************************}

  PROCEDURE tMDIDialog.wmInitMenu(VAR Msg:  tMessage);
    VAR
      CurFocus:  hWND;

    BEGIN	{ TMDIDialog.wmInitMenu }
      CurFocus := GetFocus;				{ Get the window that current has focus }
      IF (CurFocus <> 0) AND				{ If there is one that has the focus }
          IsChild(TheDialog^.hWindow, CurFocus)		{ And it's a child of the dialog }
        THEN FocusChildHandle := CurFocus;		{ Then save the child's handle }
      IF SysMenu = 0
        THEN SysMenu := GetSystemMenu(hWindow, FALSE);	{ Get the window's system menu }
      EnableMenuItem(SysMenu, sc_Maximize,		{ Disable the maximize menu item }
                     mf_Grayed OR mf_ByCommand);
      EnableMenuItem(SysMenu, sc_Size,			{ Disable the sizing   menu item }
                     mf_Grayed OR mf_ByCommand);
      DefWndProc(Msg)
    END		{ TMDIDialog.wmInitMenu };

{**********************************************************************

	This method handles the case of the TMDIDialog window being
        activated or deactivated.  The message is handled by first
        calling the TMDIChild method of the same name, so that the
        menu can be reset.  We then activate the child dialog box's
        keyboard handler and then set the focus to the correct child
        control.

**********************************************************************}

  PROCEDURE tMDIDialog.wmMDIActivate(VAR Msg:  tMessage);
    BEGIN	{ TMDIDialog.wm_MDIActivate }
      tMDIChild.wmMDIActivate(Msg);		{ Change the frame window's menu }
      ActivateDlg  (Msg);			{ Activate the dialog's keyboard handler }
      SetChildFocus(Msg)			{ Reset the focus to the correct control }
    END		{ TMDIDialog.wmMDIActivate };

{**********************************************************************

	When the TMDIDialog window is activated, it also gets the focus.
        We have to activate the child dialog box's keyboard handler and
        pass the focus to the correct dialog box child window.  No
        other processing is necessary.

**********************************************************************}

  PROCEDURE tMDIDialog.wmSetFocus(VAR Msg:  tMessage);
    BEGIN	{ TMDIDialog.wmSetFocus }
      ActivateDlg  (Msg);
      SetChildFocus(Msg);
      Msg.Result := 0
    END		{ TMDIDialog.wmSetFocus };

{**********************************************************************

	This method prevents the TMDIDialog window from being any
        larger or smaller than its child dialog box.  This helps keep
        the appearance of the window consistent with that of the
        dialog box it's impersonating on screen.

        If the window is not being displayed normally (wParam <>
        SizeNormal), we just pass this message to the default window
        procedure.  If we're displaying normally, we compute:

        The height of the non-client area;
        The height & width of the dialog;
        The height & width that the window needs to be to accomodate
        all of the above (width of dialog, height of dialog + height
        of the non-client area).

**********************************************************************}

  PROCEDURE tMDIDialog.wmSize(VAR Msg:  tMessage);
    VAR
      w, h      :  INTEGER;
      DialogRect:  tRect;

    BEGIN	{ TMDIDialog.wmSize }
      IF Msg.wParam = sizeNormal			{ If we're doing a normal show }
        THEN
          BEGIN
            WITH Attr DO
              SetWindowPos(hWindow, 0, X, Y, w, h,	{ Set the window's height without moving it }
                         swp_NoZOrder OR swp_NoMove);
            MoveWindow(TheDialog^.hWindow, 0, 0, w,	{ Correctly place the dialog in the window }
                       h - NCHeight, TRUE);
            UpdateWindow(hWindow)			{ Redraw the window immediately }
          END;
      DefWndProc(Msg)
    END		{ TMDIDialog.wmSize };

{**********************************************************************

	This method ensures that the dialog box child control focus
        is properly saved in the case where the MDI window is iconized
        but has never received wm_MDIActivate with wParam = 0.

**********************************************************************}

  PROCEDURE tMDIDialog.wmSysCommand(VAR Msg:  tMessage);
    VAR
      CurFocus:  hWND;
      MenuItem:  WORD;

    BEGIN	{ TMDIDialog.wmSysCommand }
      CurFocus := GetFocus;				{ Get the window that current has focus }
      IF (CurFocus <> 0) AND				{ If there is one that has the focus }
          IsChild(TheDialog^.hWindow, CurFocus)		{ And it's a child of the dialog }
        THEN FocusChildHandle := CurFocus;		{ Then save the child's handle }
      MenuItem := (Msg.wParam AND $FFF0);
      IF (MenuItem = sc_Maximize) OR
         (MenuItem = sc_Size)
        THEN Msg.Result := 1
        ELSE DefWndProc(Msg)
    END		{ TMDIDialog.wmSysCommand };

{**********************************************************************

	TMDIFrame methods

        This method closes & destroys a child window.  AChild is a
        pointer to an OWL object representing the child being closed.
        The method first checks the pointer to make sure it's valid,
        then it checks to see if the object's window handle is a child
        of the MDI Client window.  If both of these conditions are
        met, the window is destroyed.  Otherwise, the method just exits

**********************************************************************}

  PROCEDURE tMDIFrame.CloseChild(aChild:  pWindowsObject);
    BEGIN	{ TMDIFrame.CloseChild }
      IF (aChild <> NIL) AND
         IsChild(GetClient^.hWindow, aChild^.hWindow)
        THEN
          IF aChild^.CanClose
            THEN Dispose(aChild, Done)
    END		{ TMDIFrame.CloseChild };

{**********************************************************************

	This method is used to find the pointer to the OWL object
        which has the handle passed in TheWnd.

**********************************************************************}

  FUNCTION tMDIFrame.GetChild(TheWnd:  hWND):  pWindowsObject;

    FUNCTION HandlesMatch(Window:  pWindowsObject):  BOOLEAN; FAR;
      BEGIN	{ HandlesMatch }
        HandlesMatch := Window^.hWindow = TheWnd
      END	{ HandlesMatch };

    BEGIN	{ TMDIFrame.GetChild }
      GetChild := FirstThat(@HandlesMatch)
    END		{ TMDIFrame.GetChild };

{**********************************************************************

	This method passes the wm_InitMenu message on to the currently
        active MDI child window.  This allows us to easily implement
        child-window specific menu handling for MDI child windows,
        especially edit windows.

**********************************************************************}

  PROCEDURE tMDIFrame.wmInitMenu(VAR Msg:  tMessage);
    VAR
      CurMDIHwnd:  hWND;

    BEGIN	{ TMDIFrame.wmInitMenu }
      CurMDIHwnd := LoWord(SendMessage(ClientWnd^.hWindow, wm_MDIGetActive, 0, 0));
      IF CurMDIHwnd <> 0
        THEN SendMessage(CurMDIHwnd, wm_InitMenu, Msg.wParam, Msg.lParam);
      Msg.Result := 0
    END		{ TMDIFrame.wmInitMenu };

{**********************************************************************

	This method passes the wm_InitMenuPopup message on to the
        currently active MDI child window.  This allows us to easily
        implement child-window specific menu handling for MDI child
        windows, especially edit windows.

        The MDI child should set Msg.result to 0 if it does not wish
        the frame window to pass the message onto DefWndProc.

**********************************************************************}

  PROCEDURE tMDIFrame.wmInitMenuPopup(VAR Msg:  tMessage);
    VAR
      CurMDIHwnd:  hWND;

    BEGIN	{ TMDIFrame.wmInitMenuPopup }
      CurMDIHwnd := LoWord(SendMessage(ClientWnd^.hWindow, wm_MDIGetActive, 0, 0));
      IF CurMDIHwnd <> 0
        THEN SendMessage(CurMDIHwnd, wm_InitMenuPopup, Msg.wParam, Msg.lParam);
      Msg.Result := 0
    END		{ TMDIFrame.wmInitMenuPopup };

  BEGIN
    RegisterType(RMDIChild);
    RegisterType(RMDIDialog);
    RegisterType(RMDIFrame);
  END.
